syntax = "proto3";

import "SiLAFramework.proto";

package sila2.org.silastandard.test.unobservablepropertytest.v1;

/* This feature tests a static and a dynamic unobservable property. */
service UnobservablePropertyTest {
  /* Returns the answer to the ultimate question of life, the universe, and everything. 42. */
  rpc Get_AnswerToEverything (sila2.org.silastandard.test.unobservablepropertytest.v1.Get_AnswerToEverything_Parameters) returns (sila2.org.silastandard.test.unobservablepropertytest.v1.Get_AnswerToEverything_Responses) {}
  /* Returns the unix timestamp: The time in seconds since January 1st of 1970. */
  rpc Get_SecondsSince1970 (sila2.org.silastandard.test.unobservablepropertytest.v1.Get_SecondsSince1970_Parameters) returns (sila2.org.silastandard.test.unobservablepropertytest.v1.Get_SecondsSince1970_Responses) {}
}

/* Parameters for AnswerToEverything */
message Get_AnswerToEverything_Parameters {
}

/* Responses of AnswerToEverything */
message Get_AnswerToEverything_Responses {
  sila2.org.silastandard.Integer AnswerToEverything = 1;  /* Returns the answer to the ultimate question of life, the universe, and everything. 42. */
}

/* Parameters for SecondsSince1970 */
message Get_SecondsSince1970_Parameters {
}

/* Responses of SecondsSince1970 */
message Get_SecondsSince1970_Responses {
  sila2.org.silastandard.Integer SecondsSince1970 = 1;  /* Returns the unix timestamp: The time in seconds since January 1st of 1970. */
}
