syntax = "proto3";

package sila2.io.unitelabs.robot.gripcontroller.v1;

/* Controls a robotic gripper for e.g. labware transfer operations. */
service GripController {
  /* Closes the gripper. */
  rpc Grip (sila2.io.unitelabs.robot.gripcontroller.v1.Grip_Parameters) returns (sila2.io.unitelabs.robot.gripcontroller.v1.Grip_Responses) {}
  /* Releases the gripper. */
  rpc Release (sila2.io.unitelabs.robot.gripcontroller.v1.Release_Parameters) returns (sila2.io.unitelabs.robot.gripcontroller.v1.Release_Responses) {}
}

/* Parameters for Grip */
message Grip_Parameters {
}

/* Responses of Grip */
message Grip_Responses {
}

/* Parameters for Release */
message Release_Parameters {
}

/* Responses of Release */
message Release_Responses {
}
